/*+++
 *  title:      CommandKeys.m
 *  abstract:   Implementation of CommandKeys class for CommandKeys.preferences.
 *  author:     Tom Hageman <tom@basil.icce.rug.nl>
 *  created:    Marcg 1999
 *  modified:   (see RCS Log at end)
 *  copyleft:
 *
 *		Copyright (C) 1999  Tom R. Hageman.
 *
 *	This is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this software; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 59 Temple Place - Suite 330,
 *	Boston, MA 02111-1307, USA.
 *
 *  description:
 *      (see corresponding *.h file)
 *---*/

#ifdef RCS_ID
static const char * const RCSid = ((void)&RCSid,
	"@(#)CommandKeys.m,v 1.21 1999/04/18 09:40:00 tom Exp");
#   define RCS_CommandKeys_ID
#endif

#import "CommandKeys.h"

#import "CommandKeysItem.h"
#import "CommandKeysBrowserCell.h"
#import "CommandKeysMenuBrowserCell.h"
#import "BrowserHelper.h"
#import "DefaultsHandler.h"

#import "version.h"


#define COMMANDKEYS_NX	"NXCommandKeys"
#define COMMANDKEYS_OS	"NSCommandKeys"

#define SYSTEM_NX	"System"


static NXAtom systemOwner;

static NXBundle *bundle;


// Localized strings

#define ADD_DUPLICATE_COMMAND_ALERT_MESSAGE	NXLocalizedStringFromTableInBundle(NULL, bundle, "The command `%s' already has a key equivalent.", NULL, Add Alert message -- %s is replaced with actual command name)
#define ADD_DUPLICATE_SYSTEM_COMMAND_ALERT_MESSAGE	NXLocalizedStringFromTableInBundle(NULL, bundle, "The command `%s' is already defined in System.  Do you wish to override it for this application?", NULL, Add Alert message -- %s is replaced with actual command name)
#define ADD_DUPLICATE_KEY_ALERT_MESSAGE		NXLocalizedStringFromTableInBundle(NULL, bundle, "The key equivalent `%s' already has been used in the command(s) `%s'.", NULL, Add Alert message -- first %s is replaced with key equiv; the second with duplicate command name/names)
#define ADD_ALERT_MODIFY_BUTTON	NXLocalizedStringFromTableInBundle(NULL, bundle, "Modify", NULL, Add Alert [Modify] button label)
#define ALERT_CANCEL_BUTTON	NXLocalizedStringFromTableInBundle(NULL, bundle, "Cancel", NULL, Alert [Cancel] button label)


/* Used to store temporary information */
typedef struct {
    NXAtom owner;		// The name of the application
    HashTable *value;		// a hashTable of CommandKeysItems.
    BOOL nextstep;		// YES if NEXTSTEP default
    BOOL openstep;		// YES if OPENSTEP default (can be both).
} AppInfo;


@interface CommandKeys (Private)
- setPrefsBrowserCellTo:(AppInfo *)set inMatrix:matrix atRow:(int)row;
- (int)reloadPrefsBrowserMatrix:(Matrix *)aMatrix;
- (int)reloadDetailBrowserMatrix:(Matrix *)aMatrix;
- (BOOL)addCommand:(const char *)aCommand key:(unsigned)aKey
    inMatrix:(Matrix *)aMatrix count:(int)count;
- (void)setSelectedOwnerCommand:(const char *)aCommand key:(unsigned short)aKey;
- (void)trimOpenstepSystemDefaultsInMatrix:(Matrix *)aMatrix;
- (void)saveNextstepDefaultForCell:(CommandKeysBrowserCell *)aCell;
- (void)saveExpandedOpenstepDefaultForCell:(CommandKeysBrowserCell *)aCell;
- (void)saveAllExpandedOpenstepDefaults;
@end


@implementation CommandKeys (Private)

// Extremely simple-minded (for now) conversion functions.
static inline unsigned short KeyEquivalentFromString(const unsigned char *aString)
{
    return (aString ? *aString : 0);
}

#define KEY_EQUIV_BUF_SIZE	2

static inline const char *StringFromKeyEquivalent(char *buf, unsigned short aKey)
{
    buf[0] = aKey;
    buf[1] = '\0';
    return buf;
}

// Somewhat less simple-minded conversion functions.
static HashTable *hashTableFromCommandKeyDescription(const char *commandKeys)
{
    HashTable *result = nil;

    if (commandKeys != NULL)
    {
	const char *s = commandKeys;

	result = [[HashTable alloc] initKeyDesc:@encode(char*)];

	do
	{
	    const char *t = s, *k;
	    char *command;
	    unsigned short key;

	    // find command.
	    while (*t != ',' && *t != '\0') t++;
	    if (t == s || *t == '\0') break;
	    // Command == s[0..t-1]

	    // find key (possibly \-escaped.)
	    k = t + 1;

	    if (*k == '\\')
	    {
		// Handle possibly \-escaped comma with care.
		++k;
		if (k[0] == ',' && k[1] != ',') --k;
	    }
	    if (*k == '\0') break;
	    key = KeyEquivalentFromString(k);

	    command = alloca((int)(t - s + 1));
	    strncpy(command, s, (size_t)(t - s));
	    command[t - s] = '\0';

	    if (![result isKey:command])
	    {
		CommandKeysItem *item = [[CommandKeysItem alloc] initWithCommand:command keyEquivalent:key];

		[result insertKey:[item command] value:item];
	    }
	     
	    while (*++k && *k != ',') ;
	    s = k;
	}
	while (*s++ == ',');

	if ([result count] == 0)
	{
	    result = [result free];
	}
    }
    return result;
}

static int _compare(const void *a, const void *b)
{
    return strcasecmp(*(const char**)a, *(const char **)b);
}

// return string representation of CommandKey hashTable, sorted by command name.
// Caller is responsible for freeing its result.

const char *commandKeyDescriptionFromHashTable(HashTable *commandKeyHash)
{
    const char *result = NULL;
    int count;

    if (commandKeyHash != nil && (count = [commandKeyHash count]) > 0)
    {
	// Sort commands first.
	const char **commands = alloca(count * sizeof(const char*));
	NXHashState hashState = [commandKeyHash initState];
	const void *command;
	void *dummy;
	int i = 0;

	while ([commandKeyHash nextState:&hashState key:&command value:&dummy])
	{
	    commands[i++] = (const char *)command;
	}
	qsort(commands, count, sizeof(*commands), _compare);

      {
	NXStream *stream = NXOpenMemory(NULL, 0, NX_WRITEONLY);
	int len, maxlen;
	char *buf;

	for (i = 0;  i < count;  i++)
	{
	    const char *command = commands[i];
	    CommandKeysItem *item = [commandKeyHash valueForKey:command];
	    unsigned short key = [item keyEquivalent];
	    char keybuf[KEY_EQUIV_BUF_SIZE];

	    if (i) NXPutc(stream, ',');
	    NXPrintf(stream, "%s,%s", command, StringFromKeyEquivalent(keybuf, key));
	}
	NXPutc(stream, '\0');
	NXGetMemoryBuffer(stream, &buf, &len, &maxlen);

	// Copy string so it can be freed with free() instead of vm_deallocate()
	result = NXCopyStringBuffer(buf);

	NXCloseMemory(stream, NX_FREEBUFFER);
      }
    }
    return result;
}

- setPrefsBrowserCellTo:(AppInfo *)set inMatrix:(Matrix *)aMatrix atRow:(int)aRow
{
    id cell;

    [aMatrix insertRowAt:aRow];
    cell = [aMatrix cellAt:aRow :0];
    [cell setLoaded:YES];
    [cell setLeaf:YES];

    [cell setOwner:set->owner];
    [cell setCommandKeyHash:set->value];
    [cell setNextstep:set->nextstep];
    [cell setOpenstep:set->openstep];
    return self;
}

typedef struct {
    int row;
    id matrix;
    NXAtom globalNX;	// cached for speed.
    NXAtom globalOS;
} ScanPrefsContext;

- (int)reloadPrefsBrowserMatrix:(Matrix *)aMatrix
{
    ScanPrefsContext context;

    context.row = 0;
    context.matrix = aMatrix;
    context.globalNX = [defaultsHandler nextstepGlobalOwner];
    context.globalOS = [defaultsHandler openstepGlobalOwner];

    systemCommandKeyHash = nil;

    // Scan preferences; fill browser using callback.
    if ([defaultsHandler scanPreferences:&context] < 0) return 0;

    [self trimOpenstepSystemDefaultsInMatrix:aMatrix];

    return context.row;
}

// DefaultsHandler -scanPreferences callback.

- (BOOL)defaultsHandler:sender
    foundOwner:(NXAtom)anOwner
    nextstepValue:(const char *)nextstepValue
    openstepValue:(const char *)openstepValue
    context:(void *)aContext
{
    ScanPrefsContext *context = aContext;
    AppInfo set;
    HashTable *nextstepHash, *openstepHash;

    // Ignore global domain settings, force System to be both NS & OS.
    if (anOwner == context->globalNX) return NO;
    if (anOwner == context->globalOS) return NO;

    set.owner = anOwner;
    set.nextstep = (nextstepValue != NULL);
    set.openstep = (openstepValue != NULL);

    openstepHash = hashTableFromCommandKeyDescription(openstepValue);
    nextstepHash = hashTableFromCommandKeyDescription(nextstepValue);
    set.value = nextstepHash;

    if (openstepHash)
    {
	set.value = openstepHash;

	if (nextstepHash)
	{
	    // Merge both hashtables (OPENSTEP value takes precedence)
	    NXHashState state = [nextstepHash initState];
	    const void *command;
	    void *item;

	    while ([nextstepHash nextState:&state key:&command value:&item])
	    {
		if (![set.value isKey:command])
		{
		    // Transfer entry to result hashTable.
		    [set.value insertKey:command value:item];
		    [nextstepHash removeKey:command];
		}
	    }
	    // cleanup.
	    [nextstepHash freeObjects];
	    [nextstepHash free];
	}
    }

    if (anOwner == systemOwner)
    {
	set.openstep = set.nextstep;
	systemCommandKeyHash = set.value;
    }

    [self setPrefsBrowserCellTo:&set inMatrix:context->matrix atRow:context->row++];

    return YES;
}


- (int)reloadDetailBrowserMatrix:(Matrix *)aMatrix
{
    int count = 0;
    HashTable *commandKeyHash = [selectedOwnerCell commandKeyHash];

    if (commandKeyHash != nil)
    {
	NXHashState state = [commandKeyHash initState];
	const char *command;
	CommandKeysItem *item;

	while ([commandKeyHash nextState:&state key:(const void **)&command value:(void **)&item])
	{
	    unsigned short key = [item keyEquivalent];

	    if ([self addCommand:command key:key inMatrix:aMatrix count:count])
	    {
		count++;
	    }
	}
    }
    return count;
}

- (BOOL)addCommand:(const char *)aCommand key:(unsigned)aKey
    inMatrix:(Matrix *)aMatrix count:(int)count
{
    // Keep the list in the matrix sorted.
    // (Simple minded sort/insertion algorithm, but probably fast enough for
    //  limited number of entries (<20) expected.)
    CommandKeysMenuBrowserCell *cell;
    int i = count;

    while (--i >= 0)
    {
	const char *cellTitle = [(cell = [aMatrix cellAt:i:0]) stringValue];
	int st = strcasecmp(aCommand, cellTitle);

	if (st < 0) continue;
	if (st > 0) break;

	// case-insensitive match.  Now check case-sensitively.
	st = strcmp(aCommand, cellTitle);
	if (st < 0) continue;
	if (st > 0) break;

	// exact match -- just replace key definition.
	[cell setKeyEquivalent:aKey];
	return NO;
    }
    // Add a new entry.
    ++i;

#if DEBUG
    fprintf(stderr, "-%s (command = %s, key = %c) at %d\n", SELNAME(_cmd), aCommand, aKey, i);
#endif
    [aMatrix insertRowAt:i];
    cell = [aMatrix cellAt:i:0];
    [cell setLoaded:YES];
    [cell setLeaf:YES];

    [cell setStringValue:aCommand];
    [cell setKeyEquivalent:aKey];

    return YES;
}


- (void)setSelectedOwnerCommand:(const char *)aCommand key:(unsigned short)aKey
{
    NXBrowser *browser = [browserHelper browser];
    CommandKeysBrowserCell *cell = [browser selectedCell];

    if (cell == nil)
    {
	NXBeep();   /* Shouldn't happen. */
	return;
    }

    [cell setCommand:aCommand keyEquivalent:aKey];
    [browser displayColumn:[browser lastColumn]];

    if ([cell nextstep])
    {
	[self saveNextstepDefaultForCell:cell];
    }

    if ([cell openstep])
    {
	if ([cell owner] == systemOwner)
	{
	    systemCommandKeyHash = [cell commandKeyHash];
	    [self saveAllExpandedOpenstepDefaults];
	}
	else
	{
	    [self saveExpandedOpenstepDefaultForCell:cell];
	}
    }
}


- (void)trimOpenstepSystemDefaultsInMatrix:(Matrix *)aMatrix
{
    if (systemCommandKeyHash != nil)
    {
	int i, count = [aMatrix cellCount];

#if DEBUG
	fprintf(stderr, "-%s: system key count = %d, cell count = %d\n", SELNAME(_cmd), [systemCommandKeyHash count], count);
#endif
	for (i = 0;  i < count;  i++)
	{
	    CommandKeysBrowserCell *cell = [aMatrix cellAt:i:0];
	    HashTable *commandKeyHash = [cell commandKeyHash];

	    if (commandKeyHash && [cell openstep] && ([cell owner] != systemOwner))
	    {
		NXHashState state = [systemCommandKeyHash initState];
		const char *command;
		CommandKeysItem *sysItem;

		// Conditionally remove all key definitions that exactly match
		// a system definition (command _and_ key)
		while ([systemCommandKeyHash nextState:&state
			    key:(const void **)&command value:(void *)&sysItem])
		{
		    CommandKeysItem *item = [commandKeyHash valueForKey:command];

		    if (item && [item isEqualToCommandKeysItem:sysItem])
		    {
#if DEBUG
			fprintf(stderr, "-%s: %s removing `%s,%c'\n", SELNAME(_cmd), [cell owner], [item command], [item keyEquivalent]);
#endif
			[cell setCommand:command keyEquivalent:0];
			if ([cell commandKeyHash] == nil) break; // empty.
		    }
		}
	    }
	}
    }
}

// Support method for -saveXyzzyDefault methods below.

- (void)saveDefaultForOwner:(NXAtom)anOwner
    commandKeyHash:(HashTable *)aCommandKeyHash
    nextstep:(BOOL)nextstep
{
    const char *value = commandKeyDescriptionFromHashTable(aCommandKeyHash);

#if DEBUG
    fprintf(stderr, "saveDefault: owner = %s, value = \"%s\" (%s)\n", anOwner, value, (nextstep) ? "NS" : "OS");
#endif
    [defaultsHandler setOwner:anOwner value:value nextstep:nextstep openstep:!nextstep];
    free((void *)value);
}

- (void)saveNextstepDefaultForCell:(CommandKeysBrowserCell *)aCell
{
    [self saveDefaultForOwner:[aCell owner] commandKeyHash:[aCell commandKeyHash] nextstep:YES];
}

- (void)saveExpandedOpenstepDefaultForCell:(CommandKeysBrowserCell *)aCell
{
    HashTable *cellHash = [aCell commandKeyHash];
    HashTable *resultHash = cellHash;

    if (cellHash != nil && systemCommandKeyHash != nil)
    {
	NXHashState state;
	const void *command;
	void *item;

	resultHash = [[HashTable alloc] initKeyDesc:@encode(char*)];

	// Prime result hash with system entries (if any).
	state = [systemCommandKeyHash initState];
	while ([systemCommandKeyHash nextState:&state key:&command value:&item])
	{
	    [resultHash insertKey:command value:item];
	}

	// Now add entries from cell, possibly overwriting system definitions.
	state = [cellHash initState];
	while ([cellHash nextState:&state key:&command value:&item])
	{
	    [resultHash insertKey:command value:item];
	}
    }

    [self saveDefaultForOwner:[aCell owner] commandKeyHash:resultHash nextstep:NO];

    // free result hash (but not its contents!) if different from cell hash.
    if (resultHash != cellHash) [resultHash free];
}

- (void)saveAllExpandedOpenstepDefaults
{
    NXBrowser *browser = [browserHelper browser];
    Matrix *matrix = [browser matrixInColumn:[browser lastColumn]];
    int i, count = [matrix cellCount];

    // Write out system/global definition first.
    [self saveDefaultForOwner:systemOwner commandKeyHash:systemCommandKeyHash nextstep:NO];

    // now save all defined OPENSTEP entries (augmented with new system definition)
    for (i = 0;  i < count;  i++)
    {
	CommandKeysBrowserCell *cell = [matrix cellAt:i:0];
	
	if ([cell commandKeyHash] && [cell openstep] && ([cell owner] != systemOwner))
	{
	    [self saveExpandedOpenstepDefaultForCell:cell];
	}
    }
}

@end // CommandKeys (Private)


@implementation CommandKeys

+ initialize
{
    systemOwner = NXUniqueString(SYSTEM_NX);
    return self;
}

- init
{
    self = [super init];

    bundle = [NXBundle bundleForClass:[self class]];

    if (![NXApp loadNibForLayout:[[self class] name] owner:self]) return nil;

    [infoVersionTextField setStringValue:VERSION];

    // 'window' is an interface builder outlet.
    view = [window contentView];	// 'view' MUST be initialized
    [view removeFromSuperview];
    [window setContentView: 0];		// Don't need the window for anything
    [window free];			// So free it.

    // Browsers use custom Cell classes.
    [[browserHelper browser] setCellClass:[CommandKeysBrowserCell class]];
    [browserHelper setCellHasContentsSelector:@selector(commandKeyHash)];
    [[detailBrowserHelper browser] setCellClass:[CommandKeysMenuBrowserCell class]];
    [detailBrowserHelper setSendActionOnSelect:NO];

    return self;
}


- willSelect:sender
{
    return self;
}

- didSelect:sender
{
    // Enable relevant menu items
    [NXApp enableEdit:EDIT_ALL_ITEMS];
    [NXApp enableWindow:CLOSE_ITEM];

    // Set up defaults handler.
    if (!defaultsHandler)
    {
	defaultsHandler = [[DefaultsHandler alloc] init];
	[defaultsHandler setDelegate:self];
	[defaultsHandler setNextstepDefaultName:COMMANDKEYS_NX];
	[defaultsHandler setOpenstepDefaultName:COMMANDKEYS_OS];
    }

    // Rebuild the list of apps and preferences
    [self rescanPrefs:self];
    [browserHelper select:self];

    return self;
}


- willUnselect:sender
{
    // Make sure that we aren't editing any text fields.
    [[NXApp appWindow] endEditingFor:self];
    return self;
}

- didUnselect:sender
{
    // Hide our info panel
    [infoPanel orderOut:self];

    // Disable all Edit and Window menu items
    [NXApp enableEdit:0];
    [NXApp enableWindow:0];
    return self;
}

- didHide:sender
{
    return self;
}

- didUnhide:sender
{
///    [browserHelper select:self];
    return self;
}

- rescanPrefs:sender
{
#if DEBUG
    fprintf(stderr, "-%s\n", SELNAME(_cmd));
#endif
    selectedOwnerCell = nil;

    // Cause the app browser to reload
    [browserHelper reload:self];
    [detailBrowserHelper reload:self];

    return self;
}


// Browser delegate methods.

- (int)browser:sender fillMatrix:matrix inColumn:(int)column
{
#if DEBUG
	fprintf(stderr, "-%s (sender = %p, [browserHelper browser] = %p, [detailBrowserHelper browser] = %p)\n", SELNAME(_cmd), sender, [browserHelper browser], [detailBrowserHelper browser]);
#endif
    if (sender == [browserHelper browser])
    {
	return [self reloadPrefsBrowserMatrix:matrix];
    }
    if (sender == [detailBrowserHelper browser])
    {
	return [self reloadDetailBrowserMatrix:matrix];
    }
    return 0;
}


// Text(Field) delegate methods, + support.

- (void)_maybeEnableAddButton
{
    BOOL enable = NO;
    const char *commandString;
    unsigned short key;

    if ((key = KeyEquivalentFromString([keyEquivText stringValue])) &&
	(commandString = [commandText stringValue]) && *commandString)
    {
	Matrix *matrix = [[detailBrowserHelper browser] matrixInColumn:0];
	CommandKeysMenuBrowserCell *cell;
	const char *s;

	if (!((cell = [matrix selectedCell]) &&
	      (key == [cell keyEquivalent]) &&
	      (s = [cell stringValue]) &&
	      (strcmp(s, commandString) == 0)))
	{
	    // Check rest of matrix.
	    int i;

	    for (i = [matrix cellCount];  --i >= 0; )
	    {
		cell = [matrix cellAt:i:0];

		if (key == [cell keyEquivalent] &&
		    (s = [cell stringValue]) &&
		    (*s == *commandString) &&
		    (strcmp(s, commandString) == 0))
		{
		    break;
		}
	    }
	    if (i < 0) enable = YES;
	}
    }
    [[detailBrowserHelper addButton] setEnabled:enable];
}

- textDidGetKeys:sender isEmpty:(BOOL)isEmpty
{
    // sender == field editor, its delegate == text field.
    id theTextField = [sender delegate];

    if (theTextField == commandText)
    {
	[detailBrowserHelper textDidGetKeys:sender isEmpty:isEmpty];
    }
    [self _maybeEnableAddButton];

    return self;
}

- textDidEnd:sender endChar:(unsigned short)endChar
{
    // sender == field editor, its delegate == text field.
    id theTextField = [sender delegate];

    if (theTextField == commandText)
    {
	[detailBrowserHelper textDidEnd:sender endChar:endChar];
    }
    [self _maybeEnableAddButton];

    return self;
}


// Support methods for adding command key definitions.

- (BOOL)checkDuplicateCommand:(const char *)aCommand
{
    if ([selectedOwnerCell owner] != systemOwner &&
	[systemCommandKeyHash isKey:aCommand])
    {
	if (NXRunAlertPanel([[detailBrowserHelper addButton] title],
			    ADD_DUPLICATE_SYSTEM_COMMAND_ALERT_MESSAGE,
			    NULL,
			    ALERT_CANCEL_BUTTON,
			    NULL,
			    aCommand) != NX_ALERTDEFAULT)
	{
	    return NO;
	}
    }
    else if ([[selectedOwnerCell commandKeyHash] isKey:aCommand])
    {
	if (NXRunAlertPanel([[detailBrowserHelper addButton] title],
			    ADD_DUPLICATE_COMMAND_ALERT_MESSAGE,
			    ADD_ALERT_MODIFY_BUTTON,
			    ALERT_CANCEL_BUTTON,
			    NULL,
			    aCommand) != NX_ALERTDEFAULT)
	{
	    return NO;
	}
    }

    return YES;
}

- (void)findDuplicateKeyEquivalent:(unsigned short)aKeyEquivalent
    inCommandKeyHash:(HashTable *)aCommandKeyHash
    into:(List *)aList
{
    NXHashState state = [aCommandKeyHash initState];
    const void *command;
    CommandKeysItem *item;

    while ([aCommandKeyHash nextState:&state key:&command value:(void **)&item])
    {
	if ([item keyEquivalent] == aKeyEquivalent)
	{
	    [aList addObject:item];
	}
    }
}

- (BOOL)checkDuplicateKeyEquivalent:(const char *)aKeyEquivString
    command:(const char *)aCommand
{
    BOOL result = YES;
    unsigned short key = KeyEquivalentFromString(aKeyEquivString);
    List *duplicateList = [[List alloc] init];
    int i, count;

    [self findDuplicateKeyEquivalent:key
	inCommandKeyHash:[selectedOwnerCell commandKeyHash]
	into:duplicateList];

    if ([selectedOwnerCell owner] != systemOwner)
    {
	[self findDuplicateKeyEquivalent:key
	    inCommandKeyHash:systemCommandKeyHash
	    into:duplicateList];
    }

    if ((count = [duplicateList count]))
    {
	NXStream *stream = NXOpenMemory(NULL, 0, NX_WRITEONLY);
	int len, maxlen;
	char *duplicates;
	
	for (i = 0;  i < count;  i++)
	{
	    CommandKeysItem *item = [duplicateList objectAt:i];

	    NXPrintf(stream, (i == 0 ? "%s" : ", %s"), [item command]);
	}
	NXPutc(stream, '\0');
	NXGetMemoryBuffer(stream, &duplicates, &len, &maxlen);

	if (NXRunAlertPanel([[detailBrowserHelper addButton] title],
			    ADD_DUPLICATE_KEY_ALERT_MESSAGE,
			    ADD_ALERT_MODIFY_BUTTON,
			    ALERT_CANCEL_BUTTON,
			    NULL,
			    aKeyEquivString, duplicates) != NX_ALERTDEFAULT)
	{
	    result = NO;
	}
	NXCloseMemory(stream, NX_FREEBUFFER);
    }
    [duplicateList free];

    return result;
}


// BrowserHelper (Add, Remove) action delegates.

- (BOOL)browser:(NXBrowser *)aBrowser willAddCellNamed:(const char *)aName
    inMatrix:(Matrix *)aMatrix atRow:(int)aRow
{
    if (aBrowser == [detailBrowserHelper browser])
    {
	const char *command = aName;
	const char *keyEquivString = [keyEquivText stringValue];
	unsigned short key;

	if (!(command && *command && keyEquivString &&
	      (key = KeyEquivalentFromString(keyEquivString))))
	{
	    NXBeep();	/* Shouldn't happen... */
	    return NO;
	}

	if (![self checkDuplicateCommand:command]) return NO;
	if (![self checkDuplicateKeyEquivalent:keyEquivString command:command]) return NO;

	[self addCommand:command key:key inMatrix:aMatrix count:aRow];
	// This does the right thing even if command already is present.

	[self setSelectedOwnerCommand:command key:key];

	return NO;
    }
    return YES;
}

- (BOOL)browser:(NXBrowser *)aBrowser willRemoveCellInMatrix:(Matrix *)aMatrix
    atRow:(int)aRow
{
    if (aBrowser == [detailBrowserHelper browser])
    {
	const char *command = [[aMatrix cellAt:aRow:0] stringValue];

	[self setSelectedOwnerCommand:command key:0];
    }
    return YES;
}


// Regular action methods.

- prefSelected:sender
{
    CommandKeysBrowserCell *selectedCell = [sender selectedCell];

#if DEBUG
  {
    const char *commandKeys = commandKeyDescriptionFromHashTable( [selectedCell commandKeyHash]);
    fprintf(stderr, "-%s (selectedOwnerCell = %p %s, selectedCell = %p %s, commandKeys = %s)\n", SELNAME(_cmd), selectedOwnerCell, [selectedOwnerCell owner], selectedCell, [selectedCell owner], commandKeys);
    free((void *)commandKeys);
  }
#endif
    /* Several optimizations: reload detail browser only if owner selection
       changes, and either has command keys associated with it. (ie. don't
       reload if detail browser would stay empty. */
    if (selectedCell != selectedOwnerCell)
    {
	if ([selectedOwnerCell commandKeyHash] || [selectedCell commandKeyHash])
	{
	    selectedOwnerCell = selectedCell;
	    [detailBrowserHelper reload:sender];
	}
	else
	{
	    selectedOwnerCell = selectedCell;
	}
    }
    return self;
}

- commandKeySelected:sender
{
    CommandKeysMenuBrowserCell *cell = [sender selectedCell];

    if (cell)
    {
	char keybuf[KEY_EQUIV_BUF_SIZE];

	[commandText setStringValue:[cell stringValue]];
	[keyEquivText setStringValue:StringFromKeyEquivalent(keybuf, [cell keyEquivalent])];
	[commandText selectText:sender];

	[[detailBrowserHelper addButton] setEnabled:NO];
	[[detailBrowserHelper removeButton] setEnabled:YES];
    }
    return self;
}

@end // CommandKeys
